import os
import math
import random
import numpy as np
import pandas as pd
import pyopencl as cl
from tqdm import tqdm
from pyopencl import array as ocl_array
from multiprocessing import Pool, RLock

class GillespieSSA:
    
    def run(self):
        
        for batch_i in tqdm(range(self.nbatches), desc='Batch For Loop',leave=False):
            self.simulate(batch_i)
            
    def simulate(self, batch_number):
        # Setup Simulation Context and Parameters
        print ("Creating Command Queue...")
        context, program, cache_interval, not_perturbed, time_span = self.setup_simulation()
        
        with cl.CommandQueue(context) as queue:
            print ("Start Simulation...")
            aff_params, prop_params, change_params, species_array, species_vec, prop_df, affinities, propensities = self.initialize_variables(queue, batch_number)
            results, result_index = self.setup_results(species_vec)
            cache_species = species_vec.iloc[result_index]
            
            for time_index in tqdm(time_span, desc=f'Running Simulations Batch: {batch_number}'):
                # calculate affinites and propensities
                affinities = self.calc_affinities(program, queue, aff_params, species_array, affinities)
                propensity_vec = self.calc_propensities(program, queue, prop_params, propensities, affinities, species_array)

                # sample reactions and update species state
                k = self.get_firings(queue, propensity_vec)
                species_array = self.update_state(program, queue, species_array, results, k, change_params)

                # cache result and introduce perturbation to simulation
                not_perturbed, prop_params = self.perturb_reactions(queue, time_index, not_perturbed, prop_params, prop_df)
                cache_interval, results = self.cache_results(batch_number, time_index, cache_interval, cache_species, 
                                                             species_array, results, result_index)
            
            cache_interval, results = self.cache_results(batch_number, time_index, cache_interval, cache_species, 
                                                         species_array, results, result_index)
    
    def setup_simulation(self):
        cache_interval = 0
        not_perturbed = self.not_perturbed
        time_span = np.arange(0, self.sim_time, self.tau)
        
        context, program = self.get_gpu_context()
        
        return context, program, cache_interval, not_perturbed, time_span
    
    def setup_results(self, species_vec):
        if self.save_protein: # remove this option
            result_index = species_vec.index.values
        else:
            result_index = species_vec.loc[species_vec.molecule_type != 'protein', ].index.values
        
        result = np.zeros(shape=(self.cache_size, len(result_index)))
        return result, result_index
    
    def initialize_variables(self, queue, device_num):
        file = f'batch_{device_num}.parquet'
        aff_params, prop_params, change_params = {}, {}, {}
        affinity = pd.read_parquet(os.path.join(self.simulator_dir, 'affinity', file))
        regulators = pd.read_parquet(os.path.join(self.simulator_dir, 'regulators', file))
        propensity = pd.read_parquet(os.path.join(self.simulator_dir, 'propensity', file))
        change_vec = pd.read_parquet(os.path.join(self.simulator_dir, 'change_vec', file))
        species_vec = pd.read_parquet(os.path.join(self.simulator_dir, 'species_vec', file))
        
        # stable affinities parameters
        aff_params['hill'] = ocl_array.to_device(queue, affinity['hill'].values.astype(np.float64))
        aff_params['strengths'] = ocl_array.to_device(queue, affinity['strength'].values.astype(np.float64))
        aff_params['regulatory_index'] = ocl_array.to_device(queue, affinity['from_index'].values.astype(np.int32))
        aff_params['dissociations'] = ocl_array.to_device(queue, affinity['dissociation'].values.astype(np.float64))
        
        # stable propensities parameters
        prop_params['effects'] = ocl_array.to_device(queue, affinity['effect'].values.astype(np.int32))
        prop_params['basal'] = ocl_array.to_device(queue, propensity['basal'].values.astype(np.float64))
        prop_params['regulators'] = ocl_array.to_device(queue, regulators['regulators'].values.astype(np.int32))
        prop_params['nregulators'] = ocl_array.to_device(queue, propensity['nregulators'].values.astype(np.int32))
        prop_params['effects_sums'] = ocl_array.to_device(queue, propensity['effects_sums'].values.astype(np.int32))
        prop_params['independence'] = ocl_array.to_device(queue, propensity['independence'].values.astype(np.float64))
        prop_params['reaction_type'] = ocl_array.to_device(queue, propensity['reaction_type'].values.astype(np.int32))
        prop_params['species_index'] = ocl_array.to_device(queue, propensity['species_index'].values.astype(np.int32))
        prop_params['base_activity'] = ocl_array.to_device(queue, propensity['base_activity'].values.astype(np.float64))
        prop_params['reaction_rates'] = ocl_array.to_device(queue, propensity['reaction_rate'].values.astype(np.float64))
        prop_params['reversible_reactions'] = ocl_array.to_device(queue, propensity['reversible_reaction'].values.astype(np.int32))
        
        # stable update state parameters
        change_params['reaction_effects'] = ocl_array.to_device(queue, change_vec['reaction_effect'].values.astype(np.int32))
        change_params['change_reaction_index'] = ocl_array.to_device(queue, change_vec['react_index'].values.astype(np.int32))
        change_params['change_species_index'] = ocl_array.to_device(queue, change_vec['species_index'].values.astype(np.int32))
        
        # create species, affinities and propensities vector
        species_array = ocl_array.to_device(queue, species_vec['state'].values.astype(np.int32))
        propensities = ocl_array.zeros(queue, order='C', shape=prop_params['basal'].shape, dtype=np.float64)
        affinities = ocl_array.zeros(queue, order='C', shape=aff_params['strengths'].shape, dtype=np.float64)
        return aff_params, prop_params, change_params, species_array, species_vec, propensity, affinities, propensities
    
    def calc_affinities(self, program, queue, aff_params, species_array, affinities):
        # initialize affinities parameters
        affinities_calc = program.calc_affinities(queue,
                                                   aff_params['strengths'].shape,
                                                   None,
                                                   aff_params['strengths'].data,
                                                   aff_params['hill'].data,
                                                   aff_params['dissociations'].data,
                                                   aff_params['regulatory_index'].data,
                                                   species_array.data,
                                                   affinities.data)

        affinities_calc.wait()
        return affinities

    def calc_propensities(self, program, queue, prop_params, propensities, affinities, species_array):
        # initializing propensity variables
        propensities_calc = program.calc_propensities(queue,
                                                        prop_params['reaction_rates'].shape,
                                                        None,
                                                        prop_params['effects'].data,
                                                        prop_params['regulators'].data,
                                                        prop_params['effects_sums'].data,
                                                        prop_params['nregulators'].data,
                                                        species_array.data,
                                                        prop_params['species_index'].data,
                                                        prop_params['reaction_type'].data,
                                                        prop_params['reversible_reactions'].data,
                                                        prop_params['basal'].data,
                                                        affinities.data,
                                                        propensities.data,
                                                        prop_params['independence'].data,
                                                        prop_params['base_activity'].data,
                                                        prop_params['reaction_rates'].data)

        propensities_calc.wait()
        return propensities.get()
    
    def get_firings(self, queue, propensity_vec):
        # print (max(propensity_vec))
        k = np.random.poisson(propensity_vec * self.tau).astype(np.int32)
        return ocl_array.to_device(queue, k)
    
    def update_state(self, program, queue, species_array, results, k, change_params):
        # initialize update species data parameters
        state_calc = program.update_state(queue,
                                           change_params['reaction_effects'].shape,
                                           None,
                                           species_array.data,
                                           change_params['change_species_index'].data,
                                           change_params['change_reaction_index'].data,
                                           change_params['reaction_effects'].data,
                                           k.data)
        
        state_calc.wait()
        return species_array
    
    def perturb_reactions(self, queue, time_index, not_perturbed, prop_params, prop_df):
        
        if time_index >= self.perturb_time and not_perturbed:
            not_perturbed = False
            perturb_rates = prop_df['reaction_rate'] * prop_df['perturbation']
            prop_params['reaction_rates'] = ocl_array.to_device(queue, perturb_rates.values.astype(np.float64))
        
        return not_perturbed, prop_params
    
    def cache_results(self, device_n, time_index, cache_interval, cache_species, species_array, results, result_index):
        fp = os.path.join(self.results_dir, 'simulated_counts.csv.gz')
        meta_fp = os.path.join(self.results_dir, 'cell_metadata.csv.gz')
            
        if (time_index >= self.sample_time) and (time_index % self.update_interval == 0):
            results[cache_interval] = species_array.get()[result_index]
            
            if cache_interval == (results.shape[0] - 1):
                sims = cache_species.sim_i.unique()
                
                for sim in sims:
                    spec = cache_species.loc[cache_species.sim_i == sim,]
                    sim_res = results[:, spec.index]
                    df = pd.DataFrame(sim_res, columns=spec.spec_name.values)
                    
                    if self.collapse_mrna:
                        spec = spec.sort_values(by=['gene'])
                        mrna = spec.loc[spec.molecule_type == 'mrna', 'spec_name'].values
                        premrna = spec.loc[spec.molecule_type == 'premrna', 'spec_name'].values
                        df = pd.DataFrame(df[mrna].values + df[premrna].values, columns=mrna)
                    
                    df = df.apply(pd.to_numeric)
                    df = self.manage_dtypes(df)
                    cell_meta = self.get_cells_meta(spec, sim, df.shape[0])
                    
                    if os.path.exists(fp):
                        df.to_csv(fp, mode='a', index=False, header=False, compression='gzip')
                        cell_meta.to_csv(meta_fp, mode='a', index=False, header=False, compression='gzip')
                    else:
                        df.to_csv(fp, index=False, compression='gzip')
                        cell_meta.to_csv(meta_fp, index=False, compression='gzip')
                
                results = np.zeros(shape=(self.cache_size, len(result_index)))    
                cache_interval = 0
            else:
                cache_interval += 1
        
        return cache_interval, results
    
    def get_cells_meta(self, spec, sim_i, nrows):
        grna = spec.grna.unique()[0]
        sim_name = spec.sim_name.unique()[0]
        perturbation = spec.perturbation.sort_values().unique()[0]
        
        cell_meta = pd.DataFrame({'grna': [grna] * nrows,
                                  'sim_i': [sim_i] * nrows,
                                  'sim_name': [sim_name] * nrows,
                                  'perturbation': [perturbation] * nrows})
        
        return cell_meta
        

    def get_gpu_context(self):
        context = cl.create_some_context(True)
        
        with open(os.path.join(os.path.dirname(__file__), 'utils.cl'), 'r') as file:
            code = file.read()
            
        program = cl.Program(context, code).build(
                            options=['-cl-denorms-are-zero',
                                     '-cl-no-signed-zeros',
                                     '-cl-finite-math-only',
                                     '-cl-mad-enable'])
        return context, program



import os
import random
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from multiprocessing import Pool, RLock
from numpy.random import uniform, multinomial

class Sampling:
    
    def sample(self, ncells=10000, pop_fp=None, cache=True, sim_fp=None):
        print (f"Simulation: {self.network_name} Sampling Cells...", flush=True)
        cells_meta, gene_expr = self.sampling_cells(ncells, sim_fp)
        
        print (f"Simulation: {self.network_name} Sampling Molecules...", flush=True)
        gene_expr, lib_sizes = self.sampling_molecules(gene_expr, pop_fp)
        cells_meta = self.clean_cells_metadata(cells_meta, lib_sizes)
        
        if cache:
            print (f"Simulation: {self.network_name} Caching....", flush=True)
            cells_meta.to_csv(os.path.join(self.metadata_dir, 'cells_metadata.csv.bz2'), compression='bz2', index=False)
            gene_expr.to_csv(os.path.join(self.metadata_dir, 'gene_expression.csv.bz2'), compression='bz2', index=False)
        
        return cells_meta, gene_expr
        
    def sampling_cells(self, ncells, sim_fp):
        meta_fp = os.path.join(self.results_dir, 'cell_metadata.csv.gz')
        
        if sim_fp is None:
            sim_fp = os.path.join(self.results_dir, 'simulated_counts.csv.gz')
        
        cells_meta = pd.read_csv(meta_fp)\
                       .reset_index()\
                       .rename(columns={"index": "cell_i"})

        if ncells > cells_meta.shape[0]:
            raise Exception(f"Simulation: {self.network_name} Number of cells requested is greater than the number of cells simulated. Sample fewer cells...")
        
        cells_meta = self.sample_cells_per_grna(cells_meta, ncells)
        gene_expr = self.load_cells(cells_meta, sim_fp)
        
        return cells_meta, gene_expr
    
    def sampling_molecules(self, gene_expr, pop_fp=None):
        
        if pop_fp is None:
            pop = sc.read_loom(self.pop_fp)
        else:
            pop = sc.read_loom(pop_fp)
        
        realcounts = pop.X.toarray()
        cell_umi = pop.obs.total_counts.values
        
        lib_size = self.calc_library_size(cell_umi, gene_expr)
        simcounts_cpm = self.calc_cpm(realcounts, gene_expr)
        downsampled_counts = self.downsampling(simcounts_cpm, lib_size)
        gene_expr = pd.DataFrame(downsampled_counts, columns=gene_expr.columns, dtype=np.int16)
        
        return gene_expr, lib_size
    
    def clean_cells_metadata(self, meta, lib_sizes):
        meta['lib_size'] = lib_sizes
        meta['target_gene'] = meta['sim_name'].apply(lambda x: x.split('-grna')[0])
        
        if self.crispr_type == 'knockout':
            meta['is_cell_perturbed'] = meta['sim_name'].apply(lambda x: x.split('_')[-1])
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        else:
            meta['is_cell_perturbed'] = 'PRT'
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        
        meta = meta.reset_index(drop=True)
        return meta
    
    def load_cells(self, cells_meta, sim_fp):
        df = pd.read_csv(sim_fp, dtype=np.int16)
        df = df.iloc[cells_meta.cell_i.values]
        
        return df
        
    def calc_library_size(self, cell_umis, sim_counts):
        
        if self.map_reference_ls:
            # sampling library
            nlibs = sim_counts.shape[0]
            sim_probs = uniform(size=nlibs).astype(np.float16)
            lib_size = np.around(np.quantile(cell_umis, sim_probs))
        else:
            lib_size = sim_counts.sum(axis=1).values
        
        return lib_size
    
    def calc_cpm(self, realcount, sim_counts):
        realcount_ls = np.sum(realcount, axis=1).reshape(-1, 1)
        sim_counts_ls = sim_counts.sum(axis=1).values.reshape(-1, 1)
        
        realcount_cpm = realcount / realcount_ls
        sim_counts_cpm = sim_counts / sim_counts_ls
        
        if self.map_reference_cpm:
            # sort sim counts data via least to greatest
            sim_shape = sim_counts.shape
            realcount_cpm = realcount_cpm.flatten()
            sim_cpm_size = sim_shape[0] * sim_shape[1]
            
            realcount_cpm = realcount_cpm[realcount_cpm != 0]
            sim_probs = uniform(size=sim_cpm_size).astype(np.float16)
            sim_counts_cpm = np.quantile(realcount_cpm, sim_probs).reshape(sim_shape)
            sim_counts_cpm = sim_counts_cpm / np.sum(sim_counts_cpm, axis=1).reshape(-1, 1)
        
        return sim_counts_cpm
    
    def downsampling(self, sim_counts_cpm, lib_sizes):
        
        for index, lib_size in tqdm(enumerate(lib_sizes)):
            gene_val = sim_counts_cpm[index, ]
            gene_expr = multinomial(lib_size, gene_val).astype(np.int16)
            sim_counts_cpm[index, ] = gene_expr
        
        return sim_counts_cpm
    
    def sample_cells_per_grna(self, cells_meta, ncells):
        sampled_cells = []
        grnas = cells_meta.grna.unique()
        ncells_per_grna = round(ncells / len(grnas))
        
        for row_i in range(len(self.sim_meta)):
            row = self.sim_meta.iloc[row_i]
            sim_cells = cells_meta.loc[cells_meta.sim_name == row.sim_name, 'cell_i'].values
            sim_cells = list(sim_cells)
            
            if len(sim_cells) < ncells_per_grna:
                print ("changing ncells_per_grna...")
                ncells_per_grna = len(sim_cells)            

            if row.sample_percent != 0:
                n = int(ncells_per_grna * row.sample_percent)
                sampled = random.sample(sim_cells, k=n)
                sampled_cells = sampled_cells + sampled
        
        cells_meta = cells_meta.iloc[sampled_cells]
        return cells_meta


import os
import random
import numpy as np
import pandas as pd
import igraph as ig
from numpy.random import uniform
from .housekeeping import HouseKeeping
from .grn import GeneRegulatoryNetwork
from .modular_sampling import ModularSampling
from .signalling_cascade import SignallingCascade

    
class Network(ModularSampling, GeneRegulatoryNetwork, HouseKeeping, SignallingCascade):
    
    def create_network(self):
        
        # if precreated network is not provided create one
        if self.feature_info is None or self.feature_network is None:
            net_dfs, self.tfs, self.kinases, self.phosphorylated = [], [], [], []
            
            # load the GRN and extract the transcription factor network
            if self.verbose:
                print (f"Sampling GRN: {self.network_name}", flush=True)
            
            self.create_grn()
            
            if self.verbose:
                print (f"GRN Created: {self.network_name}", flush=True)
                
            net_dfs.append(self.get_edgelist_df(self.tf_egene_net))
            self.tfs = [f'TF_{i}' for i in range(self.num_tfs)]
            
            if self.nkinases != 0:
                if self.verbose:
                    print (f"Sampling Signalling Cascades: {self.network_name}", flush=True)
                
                self.create_signalling_cascades()
                
                if self.verbose:
                    print (f"Finished Sampling Cascades: {self.network_name}", flush=True)
                
                kinase_df = self.get_edgelist_df(self.kinases)
                self.phosphorylated = set(list(kinase_df.target.unique()) + list(kinase_df.source.unique()))
                self.kinases = [kinase for kinase in self.phosphorylated if not kinase in self.tfs]
                net_dfs.append(kinase_df)
                
            if self.num_hks != 0:
                if self.verbose:
                    print (f"Sampling HouseKeeping Network: {self.network_name}", flush=True)
                    
                self.create_hks()
                
                if self.verbose:
                    print (f"Sampled HouseKeeping Network: {self.network_name}", flush=True)
                    
                hk_df = self.get_edgelist_df(self.hks)
                self.hks = list(hk_df.source.unique()) + list(hk_df.target.unique())
                net_dfs.append(hk_df)
            
            # merged sampled GRNs
            self.merge_networks(net_dfs)
            self.create_metadata()
            print ("Created Network...", flush=True)
            
            if self.cache_network:
                self.feature_info.to_parquet(os.path.join(self.metadata_dir,
                                                          'feature_info.parquet'), 
                                             compression='brotli')
                self.feature_network.to_parquet(os.path.join(self.metadata_dir,
                                                             'feature_network.parquet'), 
                                                compression='brotli')
    
    def merge_networks(self, dfs):
        # get edgelist dataframe
        df = pd.concat(dfs)
        df = df.reset_index(drop=True)
        
        self.network = ig.Graph.DataFrame(df)
    
    def create_metadata(self):
        df_vert = self.network.get_vertex_dataframe()
        df_edge = self.network.get_edge_dataframe().reset_index(drop=True)
        
        df_vert['is_tf'] = False
        df_vert['is_hk'] = False
        df_vert['is_kinase'] = False
        df_edge['kinase_edge'] = False
        df_vert['is_phosphorylated'] = False
        
        df_edge = df_edge.drop(columns=['edge_weight'])
        df_vert = df_vert.rename(columns={'name':'feature_id'})
        df_edge = df_edge.rename(columns={'source':'from', 'target':'to'})
        df_edge['from'].replace(df_vert['feature_id'], inplace = True)
        df_edge['to'].replace(df_vert['feature_id'], inplace = True)
        
        df_vert.loc[df_vert.feature_id.isin(self.tfs), 'is_tf'] = True
        df_vert.loc[df_vert.feature_id.isin(self.hks), 'is_hk'] = True
        df_edge.loc[df_edge['from'].isin(self.kinases), 'kinase_edge'] = True
        df_vert.loc[df_vert.feature_id.isin(self.kinases), 'is_kinase'] = True
        df_vert.loc[df_vert.feature_id.isin(self.phosphorylated), 'is_phosphorylated'] = True
        
        self.feature_info = df_vert
        self.feature_network = df_edge
        
    def rename_nodes(self, network, nodes_to_rename, label):
        df_vert = network.get_vertex_dataframe()
        df_edge = network.get_edge_dataframe().reset_index(drop=True)
        nodes_to_rename = np.unique(nodes_to_rename)
        vert_loc = df_vert['name'].isin(nodes_to_rename)
                
        df_vert['name'].loc[vert_loc, ] = [f'{label}_{i}' for i in range(len(nodes_to_rename)) if nodes_to_rename[i] in df_vert['name'].values]
        df_edge['source'].replace(df_vert['name'], inplace=True)
        df_edge['target'].replace(df_vert['name'], inplace=True)
        
        network = ig.Graph.DataFrame(df_edge)
        return network
        
    def get_edgelist_df(self, network):
        df_vert = network.get_vertex_dataframe()
        df_edge = network.get_edge_dataframe().reset_index(drop=True)
        
        df_edge['source'].replace(df_vert['name'], inplace=True)
        df_edge['target'].replace(df_vert['name'], inplace=True)
        
        return df_edge